package org.briarproject.bramble.mailbox;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.mailbox.MailboxConstants;
import org.briarproject.bramble.api.sync.SyncSession;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamContext;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriter;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestEnd;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestStore;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestSync;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestTake;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.ProtocolException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.END;
import static org.briarproject.bramble.util.LogUtils.logException;

abstract class AbstractMailboxSession {
	private static final Logger LOG =
			Logger.getLogger(AbstractMailboxSession.class.getName());
	protected final Executor ioExecutor;
	protected final ContactId contactId;
	private KeyManager keyManager;
	private SyncSessionFactory syncSessionFactory;
	private final StreamWriterFactory streamWriterFactory;
	private final StreamReaderFactory streamReaderFactory;
	private MailboxProtocol mailboxProtocol;
	private final int transportMaxLatency;
	private final int transportMaxIdleTime;

	private LinkedHashMap<ContactId, PipedOutputStream>
			incomingSyncSessionWriter = new LinkedHashMap<>();

	// Used to handle graceful session termination with END request
	private AtomicBoolean remoteSessionFinished = new AtomicBoolean(false);
	protected AtomicBoolean terminated = new AtomicBoolean(false);

	public AbstractMailboxSession(Executor ioExecutor,
			KeyManager keyManager,
			SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol,
			int transportMaxLatency,
			int transportMaxIdleTime,
			ContactId contactId) {

		this.ioExecutor = ioExecutor;
		this.keyManager = keyManager;
		this.syncSessionFactory = syncSessionFactory;
		this.streamWriterFactory = streamWriterFactory;
		this.streamReaderFactory = streamReaderFactory;
		this.mailboxProtocol = mailboxProtocol;
		this.transportMaxLatency = transportMaxLatency;
		this.transportMaxIdleTime = transportMaxIdleTime;
		this.contactId = contactId;
	}

	/**
	 * @return True if the session is running
	 */
	public boolean isTerminated() {
		return terminated.get();
	}


	/**
	 * Creates and runs a duplex session which is used for sync streams between
	 * mailbox and owner (user)
	 *
	 * @return An Outgoing Duplex Session which writes MailboxProtocol SYNC requests
	 * using a MailboxSyncRequestWriter
	 * <p>
	 * Note: The stream is not encrypted, since the MailboxProtocol is responsible
	 * for encrypting/decrypting requests
	 */
	protected SyncSession createDuplexOutgoingSession() {
		MailboxSyncRequestWriter mbWriter =
				new MailboxSyncRequestWriter(mailboxProtocol);

		return syncSessionFactory
				.createDuplexOutgoingSession(contactId, transportMaxLatency,
						transportMaxIdleTime, mbWriter);
	}

	/**
	 * Get an encrypted Sync stream which can be stored on the mailbox by
	 * issuing a STORE request
	 *
	 * @param targetContactId The encrypted sync stream is generated for the given
	 * contactId
	 * @return Encrypted Sync stream
	 * @throws DbException
	 * @throws IOException
	 */
	protected byte[] getSyncStreamToStore(ContactId targetContactId)
			throws DbException, IOException {
		StreamContext ctx =
				keyManager
						.getStreamContext(targetContactId, MailboxConstants.ID);

		if (ctx == null)
			throw new IOException("Could not allocated stream context");

		ByteArrayOutputStream os = new ByteArrayOutputStream();

		StreamWriter streamWriter = streamWriterFactory
				.createStreamWriter(os, ctx);

		syncSessionFactory.createSimplexOutgoingSession(targetContactId,
				MailboxConstants.MAX_LATENCY, streamWriter);

		return os.toByteArray();
	}


	/**
	 * Must be called if the session wants to receive and handle requests
	 */
	public void readRequests() {
		while (!terminated.get()) {
			try {
				MailboxRequest req = mailboxProtocol.getNextRequest();
				if (req.getType() == END)
					return;
				ioExecutor.execute(() -> handleRequest(req));
			} catch (InterruptedException e) {
				if (LOG.isLoggable(INFO))
					LOG.info(e.toString());
			} catch (ProtocolException e) {
				if (LOG.isLoggable(INFO))
					LOG.info(e.toString());
				handleProtocolException();
			} finally {
				readingFinished();
			}
		}
	}

	/**
	 * Called when no more requests will be received from the remote
	 */
	protected void readingFinished() {
		synchronized (remoteSessionFinished) {
			remoteSessionFinished.set(true);
		}

		synchronized (incomingSyncSessionWriter) {
			//Close all Incoming SyncSessions assosciated with this MailboxSession
			for (Map.Entry<ContactId, PipedOutputStream> entry : incomingSyncSessionWriter
					.entrySet()) {
				try {
					entry.getValue().close();
				} catch (IOException e) {
					logException(LOG, WARNING, e);
				}
			}
		}
	}

	/**
	 * Calls the appropriate handler for an incoming request
	 * <p>
	 * NOTE: If a session does not support a specific request the appropriate
	 * handler needs to be overwritten. This handler should throw a
	 * {@link MailboxSessionHandleException}.
	 *
	 * @param req
	 */
	private void handleRequest(MailboxRequest req) {

		String error = null;
		try {
			switch (req.getType()) {
				case STORE:
					handleStore((MailboxRequestStore) req);
					break;
				case TAKE:
					handleTake((MailboxRequestTake) req);
					break;
				case SYNC:
					handleSync((MailboxRequestSync) req);
					break;
				default:
					throw new MailboxSessionHandleException(
							"Unsupported request type");
			}
		} catch (MailboxSessionHandleException e) {
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
			error = e.toString();
		}

		if (!req.hasResponse())
			return;

		try {
			if (error == null)
				mailboxProtocol.writeSucessResponse(req);
			else
				mailboxProtocol.writeErrorResponse(req, error);
		} catch (InterruptedException | IOException e) {
			if (LOG.isLoggable(INFO))
				LOG.info(e.toString());
		}
	}

	/**
	 * Must be called once at the end of a AbstractMailboxSession to signal the end of
	 * the session to the peer. This call blocks until the remote session
	 * signals that it was ended.
	 */
	protected void endSession()
			throws InterruptedException, IOException {

		mailboxProtocol.writeRequest(new MailboxRequestEnd());

		synchronized (remoteSessionFinished) {
			while (!remoteSessionFinished.get()) {
				remoteSessionFinished.wait();
			}
		}

		terminated.set(true);
	}


	/**
	 * Handles an incoming SYNC request. The request contains an unencrypted
	 * BSP stream which is passed to a BSP IncomingSession
	 *
	 * @param req MailboxRequestTake
	 */
	public void handleSync(MailboxRequestSync req) {
		//
		if (req.isEndOfStream())
			closeStreamForIncomingSession(contactId);

		try {
			receiveSyncStream(contactId, req.getSyncStream());
		} catch (IOException e) {
			logException(LOG, WARNING, e);
		}
	}

	private void closeStreamForIncomingSession(ContactId id) {
		synchronized (incomingSyncSessionWriter) {
			if (!incomingSyncSessionWriter.containsKey(contactId)) {
				if (LOG.isLoggable(WARNING))
					LOG.warning(
							"Received end of stream for nonexistent session");
				return;
			}

			try {
				incomingSyncSessionWriter.get(contactId).close();
			} catch (IOException e) {
				if (LOG.isLoggable(INFO))
					LOG.info(e.toString());
			}
			incomingSyncSessionWriter.remove(id);
		}
	}

	private void receiveSyncStream(ContactId id, byte[] syncStream)
			throws IOException {
		PipedOutputStream os;
		synchronized (incomingSyncSessionWriter) {
			if (!incomingSyncSessionWriter.containsKey(contactId)) {
				os = new PipedOutputStream();
				PipedInputStream is = new PipedInputStream(os);
				incomingSyncSessionWriter.put(contactId, os);
				ioExecutor.execute(() -> {
					try {
						syncSessionFactory.createIncomingSession(contactId, is)
								.run();
					} catch (IOException e) {
						if (LOG.isLoggable(INFO))
							LOG.info(e.toString());
					}
				});
			} else {
				os = incomingSyncSessionWriter.get(id);
			}
		}
		synchronized (os) {
			os.write(syncStream);
		}
	}

	/**
	 * Handles an incoming TAKE request. The request contains an encrypted BSP
	 * stream. From this stream, the tag is read and a StreamReader is created.
	 * A BSP IncomingSession is created for the stream
	 *
	 * @param req MailboxRequestTake
	 * @throws MailboxSessionHandleException
	 */
	public void handleTake(MailboxRequestTake req)
			throws MailboxSessionHandleException {
		InputStream in = new ByteArrayInputStream(req.getEncryptedSyncStream());
		byte[] tag = new byte[TAG_LENGTH];
		try {
			// read tag from input stream
			int offset = 0;
			while (offset < tag.length) {
				int read = in.read(tag, offset, tag.length - offset);
				if (read == -1) throw new EOFException();
				offset += read;
			}

			StreamContext ctx =
					keyManager.getStreamContext(MailboxConstants.ID, tag);

			if (ctx == null) {
				throw new IOException(
						"Received stream with unrecognisable tag");
			}


			InputStream reader =
					streamReaderFactory.createStreamReader(in, ctx);

			byte[] syncStream = new byte[reader.available()];
			reader.read(syncStream);
			receiveSyncStream(ctx.getContactId(), syncStream);
		} catch (DbException | IOException e) {
			throw new MailboxSessionHandleException(e.toString());
		}
	}

	public void handleStore(MailboxRequestStore req)
			throws MailboxSessionHandleException {
		// TODO: mailbox storage for Mailbox implementation
	}

	/**
	 * Called when a protocol exception occurs when reading a request
	 */
	protected abstract void handleProtocolException();

	public abstract void run() throws IOException;

	protected class MailboxSessionHandleException extends Exception {
		public MailboxSessionHandleException(String error) {
			super(error);
		}
	}
}
