package org.briarproject.bramble.mailbox;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.contact.ContactType;
import org.briarproject.bramble.api.data.BdfReader;
import org.briarproject.bramble.api.data.BdfReaderFactory;
import org.briarproject.bramble.api.data.BdfWriter;
import org.briarproject.bramble.api.data.BdfWriterFactory;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import org.briarproject.bramble.api.mailbox.MailboxConstants;
import org.briarproject.bramble.api.mailbox.MailboxInfo;
import org.briarproject.bramble.api.mailbox.MailboxManager;
import org.briarproject.bramble.api.plugin.ConnectionRegistry;
import org.briarproject.bramble.api.plugin.TransportConnectionReader;
import org.briarproject.bramble.api.plugin.TransportConnectionWriter;
import org.briarproject.bramble.api.plugin.TransportId;
import org.briarproject.bramble.api.plugin.duplex.DuplexTransportConnection;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamContext;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriter;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import javax.inject.Inject;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import static org.briarproject.bramble.util.LogUtils.logException;

public class MailboxManagerImpl implements MailboxManager {
	private static final Logger LOG =
			Logger.getLogger(MailboxManagerImpl.class.getName());
	private Executor ioExecutor;
	private MailboxSessionFactory mailboxSessionFactory;
	private ConnectionRegistry connectionRegistry;
	private StreamWriterFactory streamWriterFactory;
	private StreamReaderFactory streamReaderFactory;
	private final BdfReaderFactory bdfReaderFactory;
	private final BdfWriterFactory bdfWriterFactory;
	private KeyManager keyManager;

	private PrivateMailboxSession privateMailboxSession = null;

	@Inject
	public MailboxManagerImpl(@IoExecutor Executor ioExecutor,
			KeyManager keyManager, MailboxSessionFactory mailboxSessionFactory,
			ConnectionRegistry connectionRegistry,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			BdfReaderFactory bdfReaderFactory,
			BdfWriterFactory bdfWriterFactory) {
		this.ioExecutor = ioExecutor;
		this.keyManager = keyManager;
		this.mailboxSessionFactory = mailboxSessionFactory;
		this.connectionRegistry = connectionRegistry;
		this.streamWriterFactory = streamWriterFactory;
		this.streamReaderFactory = streamReaderFactory;
		this.bdfReaderFactory = bdfReaderFactory;
		this.bdfWriterFactory = bdfWriterFactory;
	}

	@Override
	public void handleOutgoingContactMailboxConnection(ContactId contactId,
			DuplexTransportConnection connection, TransportId transportId) {
		ioExecutor.execute(
				new ManageMailboxConnection(transportId, connection,
						ContactType.CONTACT_MAILBOX, contactId));
	}

	@Override
	public void handleOutgoingPrivateMailboxConnection(ContactId contactId,
			DuplexTransportConnection connection, TransportId transportId) {
		if (LOG.isLoggable(INFO)) {
			LOG.info("Handling Outgoing PrivateMailbox connection");
		}
		ioExecutor.execute(
				new ManageMailboxConnection(transportId, connection,
						ContactType.PRIVATE_MAILBOX, contactId));
	}

	@Override
	public void handleIncomingMailboxOwnerConnection(StreamContext ctx,
			TransportId transportId, TransportConnectionReader reader,
			TransportConnectionWriter writer) {

		ioExecutor.execute(
				new ManageMailboxConnection(ctx, transportId, reader, writer,
						ContactType.MAILBOX_OWNER));
	}

	@Override
	public void handleOwnerContactWithoutMailbox(MailboxInfo mailboxInfo) {
		if (null == privateMailboxSession ||
				privateMailboxSession.isTerminated()) {
			return;
		}

		ioExecutor.execute(
				() -> privateMailboxSession.handleOwnerContact(mailboxInfo));
	}


	private class ManageMailboxConnection implements Runnable {
		private boolean incoming;

		private final TransportId transportId;
		private final ContactType contactType;
		private final TransportConnectionWriter writer;
		private final TransportConnectionReader reader;

		private final ContactId contactId;
		private BdfWriter mailboxBdfWriter;
		private BdfReader mailboxBdfReader;
		private StreamContext incomingCtx = null;


		/**
		 * For incoming connections
		 */
		ManageMailboxConnection(StreamContext ctx,
				TransportId transportId, TransportConnectionReader reader,
				TransportConnectionWriter writer, ContactType contactType) {
			incoming = true;
			incomingCtx = ctx;
			contactId = ctx.getContactId();
			this.transportId = transportId;
			this.contactType = contactType;
			this.writer = writer;
			this.reader = reader;
		}

		/**
		 * For Outgoing connections
		 */
		public ManageMailboxConnection(TransportId transportId,
				DuplexTransportConnection connection,
				ContactType contactType, ContactId contactId) {
			this.transportId = transportId;
			this.incoming = false;
			this.contactId = contactId;
			this.contactType = contactType;
			writer = connection.getWriter();
			reader = connection.getReader();
		}

		@Override
		public void run() {
			try {
				// Create bdf writer based on (encrypted) stream writer
				// and write tag
				handleOutgoingStream();

				// Read tag from connection
				if (!incoming)
					readTag();

				// create bdf reader based on (encrypted) stream reader
				handleIncomingStream();

			} catch (IOException | DbException e) {
				logException(LOG, WARNING, e);
				disposeConnection(true);
				return;
			}

			connectionRegistry
					.registerConnection(contactId, MailboxConstants.ID,
							incoming);

			MailboxProtocol mailboxProtocol =
					new MailboxProtocol(ioExecutor, mailboxBdfWriter,
							mailboxBdfReader);

			AbstractMailboxSession mailboxSession = mailboxSessionFactory
					.createMailboxSession(mailboxProtocol, contactId,
							contactType, writer.getMaxLatency(),
							writer.getMaxIdleTime());

			ioExecutor.execute(mailboxProtocol);

			//TODO: Too hacky?
			if (contactType == ContactType.PRIVATE_MAILBOX)
				privateMailboxSession = (PrivateMailboxSession) mailboxSession;

			try {
				mailboxSession.run();
				mailboxProtocol.stop();
				disposeConnection(false);
			} catch (IOException e) {
				if (LOG.isLoggable(INFO))
					LOG.info(e.toString());
				disposeConnection(true);
			}

			connectionRegistry
					.unregisterConnection(contactId, MailboxConstants.ID,
							incoming);

		}

		private void handleOutgoingStream()
				throws DbException, IOException {
			// Allocate a stream context
			StreamContext ctx =
					keyManager.getStreamContext(contactId, transportId);

			if (ctx == null)
				throw new IOException("Unable to allocate stream context");

			StreamWriter mailboxStreamWriter =
					streamWriterFactory.createStreamWriter(
							writer.getOutputStream(), ctx);

			OutputStream os = mailboxStreamWriter.getOutputStream();

			// Write Tag to stream to allow recipient to identify the connection
			os.flush();

			mailboxBdfWriter = bdfWriterFactory.createWriter(os);
		}

		private void readTag() throws IOException, DbException {
			byte[] tag = new byte[TAG_LENGTH];
			InputStream in = reader.getInputStream();
			int offset = 0;
			while (offset < tag.length) {
				int read = in.read(tag, offset, tag.length - offset);
				if (read == -1) throw new EOFException();
				offset += read;
			}

			incomingCtx = keyManager.getStreamContext(transportId, tag);
			if (incomingCtx == null)
				throw new IOException(
						"Unrecognised tag for returning stream");

			if (!incomingCtx.getContactId().equals(contactId))
				throw new IOException(
						"ContactId for Incoming stream does not match expected ContactId");

		}

		private void handleIncomingStream() throws IOException {

			InputStream mailboxInputStream =
					streamReaderFactory.createStreamReader(
							reader.getInputStream(), incomingCtx);

			mailboxBdfReader =
					bdfReaderFactory.createReader(mailboxInputStream);
		}

		private void disposeConnection(boolean exception) {
			try {
				reader.dispose(exception, incomingCtx != null);
				writer.dispose(exception);
			} catch (IOException e) {
				logException(LOG, WARNING, e);
			}
		}

	}
}

