package org.briarproject.bramble.mailbox;

import org.briarproject.bramble.api.contact.Contact;
import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.contact.ContactType;
import org.briarproject.bramble.api.contact.event.ContactAddedEvent;
import org.briarproject.bramble.api.db.DatabaseComponent;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.db.Transaction;
import org.briarproject.bramble.api.event.Event;
import org.briarproject.bramble.api.event.EventBus;
import org.briarproject.bramble.api.event.EventListener;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import org.briarproject.bramble.api.lifecycle.ServiceException;
import org.briarproject.bramble.api.mailbox.MailboxConstants;
import org.briarproject.bramble.api.mailbox.MailboxInfo;
import org.briarproject.bramble.api.mailbox.MailboxManager;
import org.briarproject.bramble.api.mailbox.MailboxService;
import org.briarproject.bramble.api.plugin.ConnectionRegistry;
import org.briarproject.bramble.api.plugin.LanTcpConstants;
import org.briarproject.bramble.api.plugin.PluginManager;
import org.briarproject.bramble.api.plugin.TransportId;
import org.briarproject.bramble.api.plugin.duplex.DuplexPlugin;
import org.briarproject.bramble.api.plugin.duplex.DuplexTransportConnection;
import org.briarproject.bramble.api.plugin.event.TransportDisabledEvent;
import org.briarproject.bramble.api.plugin.event.TransportEnabledEvent;
import org.briarproject.bramble.api.properties.TransportProperties;
import org.briarproject.bramble.api.properties.TransportPropertyManager;
import org.briarproject.bramble.api.system.Scheduler;

import java.security.SecureRandom;
import java.util.Collection;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import javax.inject.Inject;

import static java.util.concurrent.TimeUnit.MILLISECONDS;
import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.contact.ContactType.PRIVATE_MAILBOX;
import static org.briarproject.bramble.api.mailbox.MailboxConstants.POLLING_INTERVALL;
import static org.briarproject.bramble.util.LogUtils.logException;

/**
 * The Mailbox Service runs to poll mailboxes
 * TODO/FIXME REFACTOR!
 */
public class MailboxServiceImpl implements MailboxService, EventListener {
	private static final Logger LOG =
			Logger.getLogger(MailboxServiceImpl.class.getName());

	private final Executor ioExecutor;
	private final ScheduledExecutorService scheduler;
	private EventBus eventBus;
	private final DatabaseComponent db;
	private final ConnectionRegistry connectionRegistry;
	private final PluginManager pluginManager;
	private final MailboxManager mailboxManager;
	private final TransportPropertyManager transportPropertyManager;
	private final SecureRandom random;


	private volatile Future mailboxLanFuture;
	private volatile ContactId privateMailboxId = null;
	private volatile AtomicBoolean hasPrivateMailbox = new AtomicBoolean(false);

	@Inject
	MailboxServiceImpl(@IoExecutor Executor ioExecutor, @Scheduler
			ScheduledExecutorService scheduler, EventBus eventBus,
			DatabaseComponent db,
			ConnectionRegistry connectionRegistry,
			PluginManager pluginManager, MailboxManager mailboxManager,
			TransportPropertyManager transportPropertyManager,
			SecureRandom random) {
		this.ioExecutor = ioExecutor;
		this.scheduler = scheduler;
		this.eventBus = eventBus;
		this.db = db;
		this.connectionRegistry = connectionRegistry;
		this.pluginManager = pluginManager;
		this.mailboxManager = mailboxManager;
		this.transportPropertyManager = transportPropertyManager;
		this.random = random;

	}

	@Override
	public void startService() throws ServiceException {
		if (LOG.isLoggable(INFO))
			LOG.info("Starting Mailbox Service");

		checkForPrivateMailbox();
		tryToRunLanMailboxFuture();
		this.eventBus.addListener(this);
	}

	@Override
	public void eventOccurred(Event e) {
		if (e instanceof TransportEnabledEvent) {
			if (((TransportEnabledEvent) e).getTransportId()
					.equals(LanTcpConstants.ID)) {
				if (mailboxLanFuture == null || mailboxLanFuture.isCancelled())
					tryToRunLanMailboxFuture();
			}
		}

		if (e instanceof TransportDisabledEvent) {
			if (((TransportDisabledEvent) e).getTransportId()
					.equals(LanTcpConstants.ID)) {
				if (mailboxLanFuture != null)
					mailboxLanFuture.cancel(false);
			}
		}

		if (e instanceof ContactAddedEvent) {
			if (!hasPrivateMailbox.get())
				ioExecutor.execute(() -> checkForPrivateMailbox());
		}
	}

	private void tryToRunLanMailboxFuture() {
		//NOTE/TODO: Only using LAN connections for now
		DuplexPlugin plugin =
				(DuplexPlugin) pluginManager.getPlugin(LanTcpConstants.ID);
		// TODO: Find a useful polling interval and randomize (should have
		// a minimum polling interval)
		if (plugin != null && plugin.isRunning())
			mailboxLanFuture = schedule(
					new ConnectMailboxesTask(LanTcpConstants.ID,
							POLLING_INTERVALL, plugin), POLLING_INTERVALL);
	}

	private void checkForPrivateMailbox() {
		synchronized (hasPrivateMailbox) {
			if (hasPrivateMailbox.get())
				return;

			Transaction txn = null;
			Collection<Contact> privateMb;
			try {
				txn = db.startTransaction(true);
				privateMb = db.getContactsByType(txn, PRIVATE_MAILBOX);
				db.commitTransaction(txn);

				if (privateMb.size() > 1)
					throw new RuntimeException(
							"Multiple Private Mailboxes exist!");

				if (!privateMb.isEmpty()) {
					privateMailboxId = privateMb.iterator().next().getId();
					hasPrivateMailbox.set(true);
				}
			} catch (DbException dbe) {
				logException(LOG, WARNING, dbe);
			} finally {
				if (txn != null)
					db.endTransaction(txn);
			}
		}
	}

	private Future schedule(Runnable task, int delay) {
		return scheduler
				.schedule(() -> ioExecutor.execute(task), delay, MILLISECONDS);
	}

	private class ConnectMailboxesTask implements Runnable {
		private TransportId transportId;
		private int delay;
		private DuplexPlugin plugin;
		private TransportProperties privateMailboxProperties = null;

		private ConnectMailboxesTask(TransportId transportId,
				int delay, DuplexPlugin plugin) {
			this.transportId = transportId;
			this.delay = delay;
			this.plugin = plugin;
		}

		private void reschedule() {
			mailboxLanFuture = schedule(this, delay);
		}


		@Override
		public void run() {
			if (hasPrivateMailbox.get() &&
					!connectionRegistry.isConnected(privateMailboxId)) {
				try {
					connectPrivateMailbox();
				} catch (DbException e) {
					logException(LOG, WARNING, e);
					return;
				}
			}

			// Poll Contact mailboxes
			Map<ContactId, TransportProperties> contacts;
			Collection<MailboxInfo> contactMailboxes;
			Transaction txn = null;
			try {
				txn = db.startTransaction(true);
				contactMailboxes = db.getContactMailboxes(txn);
				contacts =
						transportPropertyManager.getRemotePropertiesByType(txn,
								MailboxConstants.ID, ContactType.CONTACT);
				db.commitTransaction(txn);
			} catch (DbException e) {
				logException(LOG, WARNING, e);
				return;
			} finally {
				if (txn != null)
					db.endTransaction(txn);
			}

			for (MailboxInfo mailboxInfo : contactMailboxes) {
				if (mailboxInfo.getMailboxId() == null)
					continue;
				if (!connectionRegistry
						.isConnected(mailboxInfo.getContactId())) {
					TransportProperties mailboxTP =
							contacts.get(mailboxInfo.getMailboxId());
					if (mailboxTP == null)
						throw new AssertionError();
					ioExecutor.execute(
							() -> connectContactMailbox(plugin,
									mailboxInfo.getContactId(),
									mailboxTP));
				}
			}

			reschedule();

		}

		private void connectContactMailbox(DuplexPlugin plugin,
				ContactId contactId,
				TransportProperties transportProperties) {
			DuplexTransportConnection mailBoxConn =
					plugin.createConnection(transportProperties);
			if (mailBoxConn != null)
				mailboxManager.handleOutgoingContactMailboxConnection(
						contactId, mailBoxConn,
						transportId);
		}

		private void connectPrivateMailbox() throws DbException {
			if (privateMailboxProperties == null) {
				privateMailboxProperties = transportPropertyManager
						.getRemoteProperties(privateMailboxId, transportId);
			}

			if (LOG.isLoggable(INFO))
				LOG.info("Connecting to private mailbox");

			DuplexTransportConnection mailBoxConn =
					plugin.createConnection(privateMailboxProperties);

			if (mailBoxConn != null) {
				mailboxManager
						.handleOutgoingPrivateMailboxConnection(
								privateMailboxId, mailBoxConn,
								transportId);
			} else {
				if (LOG.isLoggable(INFO))
					LOG.info("Unable to connect to private mailbox");
			}
		}
	}

	@Override
	public void stopService() throws ServiceException {
		if (mailboxLanFuture != null && !mailboxLanFuture.isCancelled())
			mailboxLanFuture.cancel(true);
		eventBus.removeListener(this);
	}
}
