package org.briarproject.bramble.mailbox;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.mailbox.MailboxInfo;
import org.briarproject.bramble.api.sync.SyncSession;
import org.briarproject.bramble.api.sync.SyncSessionFactory;
import org.briarproject.bramble.api.transport.KeyManager;
import org.briarproject.bramble.api.transport.StreamReaderFactory;
import org.briarproject.bramble.api.transport.StreamWriterFactory;
import org.briarproject.bramble.mailbox.protocol.MailboxProtocol;
import org.briarproject.bramble.mailbox.protocol.MailboxRequest;
import org.briarproject.bramble.mailbox.protocol.MailboxRequestStore;

import java.io.IOException;
import java.net.ProtocolException;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.util.LogUtils.logException;

class PrivateMailboxSession extends AbstractMailboxSession {
	private static final Logger LOG =
			Logger.getLogger(PrivateMailboxSession.class.getName());
	private MailboxProtocol mailboxProtocol;
	private SyncSession duplexOutgoingSession;

	public PrivateMailboxSession(ContactId contactId, Executor ioExecutor,
			KeyManager keyManager,
			SyncSessionFactory syncSessionFactory,
			StreamWriterFactory streamWriterFactory,
			StreamReaderFactory streamReaderFactory,
			MailboxProtocol mailboxProtocol, int transportMaxLatency,
			int transportMaxIdleTime) {

		super(ioExecutor, keyManager, syncSessionFactory, streamWriterFactory,
				streamReaderFactory, mailboxProtocol, transportMaxLatency,
				transportMaxIdleTime, contactId);
		this.mailboxProtocol = mailboxProtocol;
	}

	/**
	 * Stores messages on the private mailbox for a contact
	 * who does not have a mailbox
	 */
	public void handleOwnerContact(MailboxInfo mailboxInfo) {
		byte[] encryptedStream;
		try {
			encryptedStream =
					getSyncStreamToStore(mailboxInfo.getContactId());
		} catch (DbException | IOException e) {
			logException(LOG, WARNING, e);
			return;
		}

		MailboxRequest req =
				new MailboxRequestStore(mailboxInfo.getAliasId(),
						encryptedStream);
		try {
			mailboxProtocol.writeRequest(req);
			if (!req.awaitAndGetResponse())
				throw new ProtocolException(req.getError());
		} catch (InterruptedException | IOException e) {
			// TODO: Handle STORE error!
			if (LOG.isLoggable(INFO))
				LOG.info(req.getError());
		}
	}

	@Override
	public void run() throws IOException {
		ioExecutor.execute(() -> super.readRequests());
		duplexOutgoingSession = createDuplexOutgoingSession();
		duplexOutgoingSession.run();
		try {
			endSession();
		} catch (InterruptedException e) {
			throw new IOException("Interrupted while ending session");
		}
	}

	@Override
	protected void handleProtocolException() {
		if (duplexOutgoingSession != null)
			duplexOutgoingSession.interrupt();
	}

	@Override
	public void handleStore(MailboxRequestStore req)
			throws MailboxSessionHandleException {
		throw new MailboxSessionHandleException(
				"Private Mailbox Session does not support the TAKE request");
	}

}
