package org.briarproject.bramble.mailbox.protocol;

import org.briarproject.bramble.api.FormatException;
import org.briarproject.bramble.api.data.BdfList;

import java.net.ProtocolException;


/**
 * MailboxRPCMessages are serialized as BdfLists and formatted as follows:
 * <p>
 * Requests (BODY depending on TYPE):
 * [TYPE, MESSAGE_ID, [BODY]]
 * [long, long [BdfList]]
 * <p>
 * Error Responses:
 * [TYPE, MESSAGE_ID, [false, ERROR_MSG]]
 * [long, long, [boolean, string]]
 * <p>
 * Successful Responses (BODY depending on TYPE):
 * [TYPE, MESSAGE_ID, [true, [BODY]]]
 * [long, long, [boolean, [BdfList]]]
 */
public interface MailboxMessage {
	static MailboxMessage parse(BdfList msg) throws ProtocolException {
		try {
			TYPE mType = TYPE.fromLong(msg.getLong(0));
			switch (mType) {
				case RESPONSE:
					return new MailboxResponse(msg);
				case SYNC:
					return new MailboxRequestSync(msg);
				case TAKE:
					return new MailboxRequestTake(msg);
				case END:
					return new MailboxRequestEnd(msg);
				case STORE:
					return new MailboxRequestStore(msg);
				default:
					throw new ProtocolException(
							"Unknown message Type received");
			}
		} catch (FormatException e) {
			throw new ProtocolException("Invalid MailboxMessageReceived");
		}
	}

	/**
	 * @return Message ID which uniquely identifies a Request and is used to
	 * match responses to a request.
	 * <p>
	 * (NOTE: This identifier is only unique for "active" requests, which
	 * wer sent and await a response. Replay protection must be provided by the
	 * underlying protocol!)
	 */
	long getId();

	/**
	 * @return true if the message expects a response
	 */
	boolean hasResponse();

	/**
	 * @return Message TYPE
	 */
	TYPE getType();

	/**
	 * Serialize the Message to a Bdf List
	 *
	 * @return BdfList which is formatted as specified for each request
	 */
	BdfList toBdfList();

	/**
	 * @param list
	 * @throws FormatException
	 */
	void parseBody(BdfList list) throws FormatException;

	enum TYPE {
		RESPONSE(0),
		END(1),
		STORE(2),
		TAKE(3),
		SYNC(4);

		private int value;

		TYPE(int value) {
			this.value = value;
		}

		public static TYPE fromLong(long b) throws IllegalArgumentException {
			if (b > TYPE.values().length) {
				throw new IllegalArgumentException(
						"Invalid Mailbox Message Type");
			}
			return TYPE.values()[(int) b];
		}

		public int getValue() {
			return value;
		}
	}
}
