package org.briarproject.bramble.mailbox.protocol;

import org.briarproject.bramble.api.data.BdfList;
import org.briarproject.bramble.api.data.BdfReader;
import org.briarproject.bramble.api.data.BdfWriter;

import java.io.EOFException;
import java.io.IOException;
import java.net.ProtocolException;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executor;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;

import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.RESPONSE;

public class MailboxProtocol implements Runnable {


	private static final Logger LOG =
			Logger.getLogger(MailboxProtocol.class.getName());

	private final Executor ioExecutor;
	private final BdfWriter mailboxBdfWriter;
	private final BdfReader mailboxBdfReader;

	volatile BlockingQueue<MailboxMessage> outQueue =
			new LinkedBlockingQueue<>();

	volatile BlockingQueue<MailboxRequest> inQueue =
			new LinkedBlockingQueue<>();

	volatile ConcurrentHashMap<Long, MailboxRequest> pendingRequests =
			new ConcurrentHashMap<>();

	private volatile AtomicBoolean stopped = new AtomicBoolean(false);
	private volatile Thread writingThread;
	private volatile Thread readingThread;


	public MailboxProtocol(Executor ioExecutor, BdfWriter mailboxBdfWriter,
			BdfReader mailboxBdfReader) {
		this.ioExecutor = ioExecutor;
		this.mailboxBdfWriter = mailboxBdfWriter;
		this.mailboxBdfReader = mailboxBdfReader;

	}

	public void sendKeepAlive() throws IOException {
		// flush the writer without pending data to send a keepalive
		if (stopped.get())
			throw new ProtocolException("Mailbox protocol has stopped");
		synchronized (mailboxBdfWriter) {
			mailboxBdfWriter.flush();
		}
	}

	public void writeRequest(MailboxMessage req)
			throws InterruptedException, IOException {
		if (req.hasResponse())
			pendingRequests.put(req.getId(), (MailboxRequest) req);
		writeMailboxMessage(req);

	}

	public void writeSucessResponse(MailboxRequest req)
			throws InterruptedException, IOException {
		if (!req.hasResponse())
			throw new RuntimeException(
					"Trying to send a response for a request that does not expect a response");
		writeMailboxMessage(req.createSuccessResponse());
	}

	public void writeErrorResponse(MailboxRequest req, String error)
			throws InterruptedException, IOException {
		if (!req.hasResponse())
			throw new RuntimeException(
					"Trying to send a response for a request that does not expect a response");

		writeMailboxMessage(req.createErrorResponse(error));
	}

	private void writeMailboxMessage(MailboxMessage msg)
			throws InterruptedException, ProtocolException {
		synchronized (outQueue) {
			if (stopped.get())
				throw new ProtocolException("Protocol has stopped");
			outQueue.put(msg);
		}
	}

	public MailboxRequest getNextRequest()
			throws InterruptedException, ProtocolException {
		MailboxRequest req =  inQueue.take();
		if (stopped.get())
			throw new ProtocolException("Protocol has stopped");

		return req;
	}

	@Override
	public void run() {
		ioExecutor.execute(() -> writeOutgoingMessages());
		readIncomingMessages();
	}

	/**
	 * Call once the protocol should terminate, before the connection is closed
	 */
	public void stop() {
		if (!stopped.compareAndSet(false, true))
			throw new RuntimeException("Stopping already stopped protocol");
	}

	private void readIncomingMessages() {
		readingThread = Thread.currentThread();
		BdfList bdfMsg;
		MailboxMessage mailboxMessage;

		while (!stopped.get()) {
			try {
				if (mailboxBdfReader.eof())
					throw new EOFException();
				bdfMsg = mailboxBdfReader.readList();
			} catch (IOException e) {
				readingThread.interrupt();
				handleIOException(true, e);
				return;
			}

			try {
				mailboxMessage = MailboxMessage.parse(bdfMsg);
			} catch (ProtocolException e) {
				LOG.info(e.toString());
				continue;
			}

			if (mailboxMessage.getType() == RESPONSE)
				handleResponse((MailboxResponse) mailboxMessage);
			else
				inQueue.add((MailboxRequest) mailboxMessage);
		}
	}

	private void handleResponse(MailboxResponse msg) {
		MailboxRequest req = pendingRequests.remove(msg.getId());

		if (req == null) {
			if (LOG.isLoggable(WARNING))
				LOG.warning("Received response not matching with any request");
			return;
		}

		if (msg.isSuccess())
			req.signalSucess();
		else
			req.signalError(msg.getErrorMessage());

	}

	private void writeOutgoingMessages() {
		writingThread = Thread.currentThread();
		MailboxMessage message;

		while (!stopped.get()) {
			try {
				message = outQueue.take();
			} catch (InterruptedException e) {
				writingThread.interrupt();
				handleIOException(false,
						new IOException("Writer Interrupted"));
				return;
			}

			synchronized (mailboxBdfWriter) {
				try {
					mailboxBdfWriter.writeList(message.toBdfList());
					mailboxBdfWriter.flush();
				} catch (IOException e) {
					writingThread.interrupt();
					handleIOException(false, e);
					return;
				}
			}
		}
	}

	private void handleIOException(boolean isReadingThread, IOException e) {
		if (isReadingThread && writingThread != null)
			writingThread.interrupt();

		// Exception occurred after the protocol was stopped externally
		if (stopped.getAndSet(true))
			return;

		if (LOG.isLoggable(INFO))
			LOG.info("Mailbox Protocol was closed: " + e.toString());

		// Signal that the session has ended to a thread waiting for requests
		inQueue.clear();
		inQueue.add(new MailboxRequestEnd());

		// Ensure that threads waiting to write a request terminate
		outQueue.clear();

		// Signal the error to anyone waiting for a response
		for (Entry<Long, MailboxRequest> entry : pendingRequests.entrySet()) {
			MailboxResponse r = new MailboxResponse(entry.getKey(), false,
					"Connection closed");
			entry.getValue().signalError(r.getErrorMessage());
		}
	}


}
