package org.briarproject.bramble.mailbox.protocol;

import org.briarproject.bramble.api.FormatException;
import org.briarproject.bramble.api.data.BdfList;

import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;


public abstract class MailboxRequest implements MailboxMessage {

	protected long msgId;
	private TYPE type;
	private String error;
	private AtomicBoolean responseReceived = new AtomicBoolean(false);
	private boolean wasSuccessful = false;

	static AtomicLong msgIdCounter = new AtomicLong();

	public MailboxRequest(TYPE type) {
		this.type = type;
		this.msgId = msgIdCounter.getAndIncrement();
	}

	public MailboxRequest(BdfList lst) throws FormatException {
		this.type = TYPE.fromLong(lst.getLong(0));
		this.msgId = lst.getLong(1);
		parseBody(lst.getList(2));
	}

	@Override
	public BdfList toBdfList() {
		return BdfList.of(type.getValue(), getId(), getRequestBody());
	}

	@Override
	public long getId() {
		return msgId;
	}

	@Override
	public TYPE getType() {
		return type;
	}


	public void signalSucess() {
		synchronized (responseReceived) {
			responseReceived.set(true);
			wasSuccessful = true;
			responseReceived.notifyAll();
		}
	}

	public void signalError(String error) {
		synchronized (responseReceived) {
			responseReceived.set(true);
			wasSuccessful = false;
			this.error = error;
			responseReceived.notifyAll();
		}
	}

	/**
	 * Blocks until a response for this request has been received
	 *
	 * @return true if response indicates success, false if not
	 * @throws InterruptedException
	 */
	public boolean awaitAndGetResponse() throws InterruptedException {
		if (!hasResponse())
			throw new RuntimeException(
					"Attempting to wait for a response of a request with no response");

		synchronized (responseReceived) {
			while (!responseReceived.get())
				responseReceived.wait();
		}

		return wasSuccessful;
	}

	public String getError() {
		if (!responseReceived.get() ||
				(responseReceived.get() && wasSuccessful))
			throw new RuntimeException(
					"Trying to get Error from unfinished or successful request");
		return error;
	}

	public MailboxResponse createSuccessResponse() {
		return new MailboxResponse(msgId, true, null);
	}

	public MailboxResponse createErrorResponse(String error) {
		return new MailboxResponse(msgId, false, error);
	}

	protected abstract BdfList getRequestBody();
}

