package org.briarproject.bramble.mailbox.protocol;

import org.briarproject.bramble.api.contact.ContactId;
import org.briarproject.bramble.api.data.BdfReader;
import org.briarproject.bramble.api.data.BdfReaderFactory;
import org.briarproject.bramble.api.data.BdfWriter;
import org.briarproject.bramble.api.data.BdfWriterFactory;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import org.briarproject.bramble.test.BrambleTestCase;
import org.junit.Before;
import org.junit.Test;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.ProtocolException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executor;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;

import javax.inject.Inject;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.briarproject.bramble.mailbox.protocol.MailboxMessage.TYPE.END;
import static org.junit.Assert.assertEquals;

public class MailboxProtocolIntegrationTest extends BrambleTestCase {

	@IoExecutor
	Executor ioExecutor;

	@Inject
	BdfReaderFactory bdfReaderFactory;

	@Inject
	BdfWriterFactory bdfWriterFactory;


	private BdfWriter bdfWriter;
	private BdfReader bdfReader;
	private PipedOutputStream pipedOS;
	private PipedInputStream pipedIS;

	private MailboxProtocol mailboxProtocol;


	public MailboxProtocolIntegrationTest() {
		// The thread pool is unbounded, so use direct handoff
		BlockingQueue<Runnable> queue = new SynchronousQueue<>();
		// Discard tasks that are submitted during shutdown
		RejectedExecutionHandler policy =
				new ThreadPoolExecutor.DiscardPolicy();
		// Create threads as required and keep them in the pool for 60 seconds
		ioExecutor = new ThreadPoolExecutor(0, Integer.MAX_VALUE,
				60, SECONDS, queue, policy);

		MailboxProtocolIntegrationTestComponent component =
				DaggerMailboxProtocolIntegrationTestComponent.builder().build();
		component.inject(this);

	}

	@Before
	public void initialize() throws IOException {

		pipedIS = new PipedInputStream();
		pipedOS = new PipedOutputStream(pipedIS);


		bdfWriter = bdfWriterFactory.createWriter(pipedOS);
		bdfReader = bdfReaderFactory.createReader(pipedIS);
		mailboxProtocol = new MailboxProtocol(ioExecutor, bdfWriter, bdfReader);
	}

	@Test(expected = ProtocolException.class)
	public void testProtocolTerminationOnClosedConnection()
			throws IOException, InterruptedException {
		ioExecutor.execute(mailboxProtocol);
		MailboxRequest req =
				new MailboxRequestStore(new ContactId(123), "test".getBytes());
		mailboxProtocol.writeRequest(req);

		// read request to ensure writing thread is running
		mailboxProtocol.getNextRequest();

		// close "connection"
		pipedOS.close();

		// And "End" request should be the next read request
		try {
			MailboxRequest recvReq = mailboxProtocol.getNextRequest();
		}finally {
			// pending request should be marked as failed
			assertEquals(false, req.awaitAndGetResponse());
			assertEquals("Connection closed", req.getError());
		}
	}


	@Test
	public void testRequestsAndHandling()
			throws IOException, InterruptedException {
		ioExecutor.execute(mailboxProtocol);
		try {
			errorResponse();
			storeRequest();
			syncRequest();
			takeRequest();
			endRequest();
		} finally {
			mailboxProtocol.stop();
			pipedOS.close();
		}
	}

	private void errorResponse() throws InterruptedException, IOException {
		// Test Error Response
		MailboxRequestTake req =
				new MailboxRequestTake(new ContactId(123), "Test".getBytes());
		mailboxProtocol.writeRequest(req);
		MailboxRequestTake recvReq =
				(MailboxRequestTake) mailboxProtocol.getNextRequest();
		mailboxProtocol.writeErrorResponse(recvReq, "Error");

		assertEquals(false, req.awaitAndGetResponse());
		assertEquals("Error", req.getError());
	}

	private void takeRequest() throws IOException, InterruptedException {
		// Generate and write TAKE request
		MailboxRequestTake req =
				new MailboxRequestTake(new ContactId(123), "Test".getBytes());
		mailboxProtocol.writeRequest(req);

		MailboxRequestTake recvReq =
				(MailboxRequestTake) mailboxProtocol.getNextRequest();

		assertEquals(req.getId(), recvReq.getId());
		assertEquals(req.getType(), recvReq.getType());
		assertEquals(req.getContactId(), recvReq.getContactId());
		assertEquals("Test", new String(req.getEncryptedSyncStream()));

		// Write Success Response
		mailboxProtocol.writeSucessResponse(recvReq);

		assertEquals(true, req.awaitAndGetResponse());
	}

	private void endRequest() throws IOException, InterruptedException {
		MailboxRequest req = new MailboxRequestEnd();
		mailboxProtocol.writeRequest(req);
		MailboxRequest recvReq = mailboxProtocol.getNextRequest();
		assertEquals(req.getId(), recvReq.getId());
		assertEquals(req.getType(), recvReq.getType());
	}


	private void storeRequest() throws IOException, InterruptedException {
		// Generate and write TAKE request
		MailboxRequestStore req =
				new MailboxRequestStore(new ContactId(123), "Test".getBytes());
		mailboxProtocol.writeRequest(req);

		MailboxRequestStore recvReq =
				(MailboxRequestStore) mailboxProtocol.getNextRequest();

		assertEquals(req.getId(), recvReq.getId());
		assertEquals(req.getType(), recvReq.getType());
		assertEquals(req.getContactId(), recvReq.getContactId());
		assertEquals("Test", new String(req.getEncryptedSyncStream()));

		// Write Success Response
		mailboxProtocol.writeSucessResponse(recvReq);
		assertEquals(true, req.awaitAndGetResponse());
	}

	private void syncRequest() throws IOException, InterruptedException {
		MailboxRequestSync req =
				new MailboxRequestSync("Test".getBytes(), false);
		mailboxProtocol.writeRequest(req);
		MailboxRequestSync recvReq =
				(MailboxRequestSync) mailboxProtocol.getNextRequest();
		assertEquals(req.getId(), recvReq.getId());
		assertEquals(req.getType(), recvReq.getType());
		assertEquals("Test", new String(recvReq.getSyncStream()));
		assertEquals(false, recvReq.isEndOfStream());
	}


}