package org.briarproject.mailbox.android.settings;

import android.annotation.TargetApi;
import android.app.AlertDialog;
import android.content.Context;
import android.content.Intent;
import android.graphics.drawable.ColorDrawable;
import android.os.Bundle;
import android.support.annotation.StringRes;
import android.support.v14.preference.SwitchPreference;
import android.support.v4.content.ContextCompat;
import android.support.v4.text.TextUtilsCompat;
import android.support.v7.preference.ListPreference;
import android.support.v7.preference.Preference;
import android.support.v7.preference.Preference.OnPreferenceChangeListener;
import android.support.v7.preference.PreferenceFragmentCompat;
import android.support.v7.preference.PreferenceGroup;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import org.briarproject.bramble.api.db.DbException;
import org.briarproject.bramble.api.event.Event;
import org.briarproject.bramble.api.event.EventBus;
import org.briarproject.bramble.api.event.EventListener;
import org.briarproject.bramble.api.nullsafety.MethodsNotNullByDefault;
import org.briarproject.bramble.api.nullsafety.ParametersNotNullByDefault;
import org.briarproject.bramble.api.plugin.BluetoothConstants;
import org.briarproject.bramble.api.plugin.TorConstants;
import org.briarproject.bramble.api.settings.Settings;
import org.briarproject.bramble.api.settings.SettingsManager;
import org.briarproject.bramble.api.settings.event.SettingsUpdatedEvent;
import org.briarproject.bramble.api.system.AndroidExecutor;
import org.briarproject.bramble.api.system.LocationUtils;
import org.briarproject.bramble.plugin.tor.CircumventionProvider;
import org.briarproject.mailbox.R;
import org.briarproject.mailbox.android.Localizer;
import org.briarproject.mailbox.android.navdrawer.NavDrawerActivity;
import org.briarproject.mailbox.android.util.UiUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.logging.Logger;

import javax.inject.Inject;

import static android.content.Intent.FLAG_ACTIVITY_CLEAR_TASK;
import static android.content.Intent.FLAG_ACTIVITY_NEW_TASK;
import static android.os.Build.VERSION.SDK_INT;
import static android.provider.Settings.ACTION_CHANNEL_NOTIFICATION_SETTINGS;
import static android.provider.Settings.EXTRA_APP_PACKAGE;
import static android.provider.Settings.EXTRA_CHANNEL_ID;
import static android.support.v4.view.ViewCompat.LAYOUT_DIRECTION_LTR;
import static java.util.logging.Level.INFO;
import static java.util.logging.Level.WARNING;
import static org.briarproject.bramble.api.plugin.BluetoothConstants.PREF_BT_ENABLE;
import static org.briarproject.bramble.api.plugin.TorConstants.PREF_TOR_MOBILE;
import static org.briarproject.bramble.api.plugin.TorConstants.PREF_TOR_NETWORK;
import static org.briarproject.bramble.api.plugin.TorConstants.PREF_TOR_NETWORK_AUTOMATIC;
import static org.briarproject.bramble.util.LogUtils.logDuration;
import static org.briarproject.bramble.util.LogUtils.logException;
import static org.briarproject.bramble.util.LogUtils.now;
import static org.briarproject.mailbox.android.TestingConstants.FEATURE_FLAG_DARK_THEME;
import static org.briarproject.mailbox.android.TestingConstants.FEATURE_FLAG_PIN_LOCK;
import static org.briarproject.mailbox.android.TestingConstants.FEATURE_FLAG_SIGN_IN_REMINDER;
import static org.briarproject.mailbox.android.TestingConstants.IS_DEBUG_BUILD;
import static org.briarproject.mailbox.android.navdrawer.NavDrawerActivity.INTENT_SIGN_OUT;
import static org.briarproject.mailbox.android.util.UiUtils.hasScreenLock;
import static org.briarproject.mailbox.android.util.UiUtils.triggerFeedback;
import static org.briarproject.mailbox.api.android.AndroidNotificationManager.PREF_NOTIFY_LOCK_SCREEN;

@MethodsNotNullByDefault
@ParametersNotNullByDefault
public class SettingsFragment extends PreferenceFragmentCompat
		implements EventListener, OnPreferenceChangeListener {

	public static final String SETTINGS_NAMESPACE = "android-ui";
	public static final String BT_NAMESPACE = BluetoothConstants.ID.getString();
	public static final String TOR_NAMESPACE = TorConstants.ID.getString();
	public static final String LANGUAGE = "pref_key_language";
	public static final String PREF_SCREEN_LOCK = "pref_key_lock";
	public static final String PREF_SCREEN_LOCK_TIMEOUT =
			"pref_key_lock_timeout";
	public static final String NOTIFY_SIGN_IN = "pref_key_notify_sign_in";
	public static final String TOR_NETWORK = "pref_key_tor_network";
	public static final String TOR_MOBILE = "pref_key_tor_mobile_data";

	private static final Logger LOG =
			Logger.getLogger(SettingsFragment.class.getName());

	private SettingsActivity listener;
	private ListPreference language;
	private ListPreference enableBluetooth;
	private ListPreference torNetwork;
	private SwitchPreference torMobile;
	private SwitchPreference screenLock;
	private ListPreference screenLockTimeout;
	private SwitchPreference notifyLockscreen;

	private Preference notifySound;

	// Fields that are accessed from background threads must be volatile
	volatile Settings settings;
	@Inject
	volatile SettingsManager settingsManager;
	@Inject
	volatile EventBus eventBus;
	@Inject
	LocationUtils locationUtils;
	@Inject
	CircumventionProvider circumventionProvider;

	@Inject
	AndroidExecutor androidExecutor;

	@Override
	public void onAttach(Context context) {
		super.onAttach(context);
		listener = (SettingsActivity) context;
		// we need to inject here,
		// because onActivityCreated() is called after onCreatePreferences()
		listener.getActivityComponent().inject(this);
	}

	@Override
	public void onCreatePreferences(Bundle bundle, String s) {
		addPreferencesFromResource(R.xml.settings);

		language = (ListPreference) findPreference(LANGUAGE);
		setLanguageEntries();
		ListPreference theme =
				(ListPreference) findPreference("pref_key_theme");
		enableBluetooth = (ListPreference) findPreference("pref_key_bluetooth");
		torNetwork = (ListPreference) findPreference(TOR_NETWORK);
		torMobile = (SwitchPreference) findPreference(TOR_MOBILE);
		SwitchPreference notifySignIn =
				(SwitchPreference) findPreference(NOTIFY_SIGN_IN);
		screenLock = (SwitchPreference) findPreference(PREF_SCREEN_LOCK);
		screenLockTimeout =
				(ListPreference) findPreference(PREF_SCREEN_LOCK_TIMEOUT);
		notifyLockscreen = (SwitchPreference) findPreference(
				"pref_key_notify_lock_screen");

		language.setOnPreferenceChangeListener(this);
		theme.setOnPreferenceChangeListener((preference, newValue) -> {
			if (getActivity() != null) {
				// activate new theme
				UiUtils.setTheme(getActivity(), (String) newValue);
				// bring up parent activity, so it can change its theme as well
				// upstream bug: https://issuetracker.google.com/issues/38352704
				Intent intent =
						new Intent(getActivity(), NavDrawerActivity.class);
				intent.setFlags(
						FLAG_ACTIVITY_CLEAR_TASK | FLAG_ACTIVITY_NEW_TASK);
				startActivity(intent);
				// bring this activity back to the foreground
				intent = new Intent(getActivity(), getActivity().getClass());
				startActivity(intent);
				getActivity().finish();
			}
			return true;
		});
		enableBluetooth.setOnPreferenceChangeListener(this);
		torNetwork.setOnPreferenceChangeListener(this);
		torMobile.setOnPreferenceChangeListener(this);
		screenLock.setOnPreferenceChangeListener(this);
		screenLockTimeout.setOnPreferenceChangeListener(this);
		if (SDK_INT >= 21) {
			notifyLockscreen.setVisible(true);
			notifyLockscreen.setOnPreferenceChangeListener(this);
		}

		findPreference("pref_key_send_feedback").setOnPreferenceClickListener(
				preference -> {
					triggerFeedback(androidExecutor);
					return true;
				});

		if (IS_DEBUG_BUILD) {
			findPreference("pref_key_explode").setOnPreferenceClickListener(
					preference -> {
						throw new RuntimeException("Boom!");
					}
			);
		} else {
			theme.setVisible(FEATURE_FLAG_DARK_THEME);
			notifySignIn.setVisible(FEATURE_FLAG_SIGN_IN_REMINDER);
			screenLock.setVisible(FEATURE_FLAG_PIN_LOCK);
			screenLockTimeout.setVisible(FEATURE_FLAG_PIN_LOCK);

			findPreference("pref_key_explode").setVisible(false);
			findPreference("pref_key_test_data").setVisible(false);
			PreferenceGroup testing =
					findPreference("pref_key_explode").getParent();
			if (testing == null) throw new AssertionError();
			testing.setVisible(false);
		}

	}

	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container,
			Bundle savedInstanceState) {
		View view = super.onCreateView(inflater, container, savedInstanceState);
		ColorDrawable divider = new ColorDrawable(
				ContextCompat.getColor(getContext(), R.color.divider));
		setDivider(divider);
		return view;
	}

	@Override
	public void onStart() {
		super.onStart();
		eventBus.addListener(this);
		setSettingsEnabled(false);
		loadSettings();
	}

	@Override
	public void onStop() {
		super.onStop();
		eventBus.removeListener(this);
	}

	private void setLanguageEntries() {
		CharSequence[] tags = language.getEntryValues();
		List<CharSequence> entries = new ArrayList<>(tags.length);
		List<CharSequence> entryValues = new ArrayList<>(tags.length);
		for (CharSequence cs : tags) {
			String tag = cs.toString();
			if (tag.equals("default")) {
				entries.add(getString(R.string.pref_language_default));
				entryValues.add(tag);
				continue;
			}
			Locale locale = Localizer.getLocaleFromTag(tag);
			if (locale == null)
				throw new IllegalStateException();
			// Exclude RTL locales on API < 17, they won't be laid out correctly
			if (SDK_INT < 17 && !isLeftToRight(locale)) {
				if (LOG.isLoggable(INFO))
					LOG.info("Skipping RTL locale " + tag);
				continue;
			}
			String nativeName = locale.getDisplayName(locale);
			// Fallback to English if the name is unknown in both native and
			// current locale.
			if (nativeName.equals(tag)) {
				String tmp = locale.getDisplayLanguage(Locale.ENGLISH);
				if (!tmp.isEmpty() && !tmp.equals(nativeName))
					nativeName = tmp;
			}
			// Prefix with LRM marker to prevent any RTL direction
			entries.add("\u200E" + nativeName.substring(0, 1).toUpperCase()
					+ nativeName.substring(1));
			entryValues.add(tag);
		}
		language.setEntries(entries.toArray(new CharSequence[0]));
		language.setEntryValues(entryValues.toArray(new CharSequence[0]));
	}

	private boolean isLeftToRight(Locale locale) {
		// TextUtilsCompat returns the wrong direction for Hebrew on some phones
		String language = locale.getLanguage();
		if (language.equals("iw") || language.equals("he")) return false;
		int direction = TextUtilsCompat.getLayoutDirectionFromLocale(locale);
		return direction == LAYOUT_DIRECTION_LTR;
	}

	private void setTorNetworkSummary(int torNetworkSetting) {
		if (torNetworkSetting != PREF_TOR_NETWORK_AUTOMATIC) {
			torNetwork.setSummary("%s");  // use setting value
			return;
		}

		// Look up country name in the user's chosen language if available
		String country = locationUtils.getCurrentCountry();
		String countryName = country;
		for (Locale locale : Locale.getAvailableLocales()) {
			if (locale.getCountry().equalsIgnoreCase(country)) {
				countryName = locale.getDisplayCountry();
				break;
			}
		}
		boolean blocked =
				circumventionProvider.isTorProbablyBlocked(country);
		boolean useBridges = circumventionProvider.doBridgesWork(country);
		String setting =
				getString(R.string.tor_network_setting_without_bridges);
		if (blocked && useBridges) {
			setting = getString(R.string.tor_network_setting_with_bridges);
		} else if (blocked) {
			setting = getString(R.string.tor_network_setting_never);
		}
		torNetwork.setSummary(
				getString(R.string.tor_network_setting_summary, setting,
						countryName));
	}

	private void loadSettings() {
		listener.runOnDbThread(() -> {
			try {
				long start = now();
				settings = settingsManager.getSettings(SETTINGS_NAMESPACE);
				Settings btSettings = settingsManager.getSettings(BT_NAMESPACE);
				Settings torSettings =
						settingsManager.getSettings(TOR_NAMESPACE);
				logDuration(LOG, "Loading settings", start);
				boolean btSetting =
						btSettings.getBoolean(PREF_BT_ENABLE, false);
				int torNetworkSetting = torSettings.getInt(PREF_TOR_NETWORK,
						PREF_TOR_NETWORK_AUTOMATIC);
				boolean torMobileSetting =
						torSettings.getBoolean(PREF_TOR_MOBILE, true);
				displaySettings(btSetting, torNetworkSetting, torMobileSetting);
			} catch (DbException e) {
				logException(LOG, WARNING, e);
			}
		});
	}

	private void displaySettings(boolean btSetting, int torNetworkSetting,
			boolean torMobileSetting) {
		listener.runOnUiThreadUnlessDestroyed(() -> {
			enableBluetooth.setValue(Boolean.toString(btSetting));
			torNetwork.setValue(Integer.toString(torNetworkSetting));
			setTorNetworkSummary(torNetworkSetting);
			torMobile.setChecked(torMobileSetting);
			displayScreenLockSetting();

			if (SDK_INT < 26) {
				notifyLockscreen.setChecked(settings.getBoolean(
						PREF_NOTIFY_LOCK_SCREEN, false));
			} else {
				notifyLockscreen.setVisible(false);
			}
			setSettingsEnabled(true);
		});
	}

	private void setSettingsEnabled(boolean enabled) {
		// preferences not needed here, because handled by SharedPreferences:
		// - pref_key_theme
		// - pref_key_notify_sign_in
		// preferences partly needed here, because they have their own logic
		// - pref_key_lock (screenLock -> displayScreenLockSetting())
		// - pref_key_lock_timeout (screenLockTimeout)
		enableBluetooth.setEnabled(enabled);
		torNetwork.setEnabled(enabled);
		torMobile.setEnabled(enabled);
		if (!enabled) screenLock.setEnabled(false);
		notifyLockscreen.setEnabled(enabled);
	}

	private void displayScreenLockSetting() {
		if (SDK_INT < 21) {
			screenLock.setVisible(false);
			screenLockTimeout.setVisible(false);
		} else {
			if (getActivity() != null && hasScreenLock(getActivity())) {
				screenLock.setEnabled(true);
				screenLock.setChecked(
						settings.getBoolean(PREF_SCREEN_LOCK, false));
				screenLock.setSummary(R.string.pref_lock_summary);
			} else {
				screenLock.setEnabled(false);
				screenLock.setChecked(false);
				screenLock.setSummary(R.string.pref_lock_disabled_summary);
			}
			// timeout depends on screenLock and gets disabled automatically
			int timeout = settings.getInt(PREF_SCREEN_LOCK_TIMEOUT,
					Integer.valueOf(getString(
							R.string.pref_lock_timeout_value_default)));
			String newValue = String.valueOf(timeout);
			screenLockTimeout.setValue(newValue);
			setScreenLockTimeoutSummary(newValue);
		}
	}

	private void setScreenLockTimeoutSummary(String timeout) {
		String never = getString(R.string.pref_lock_timeout_value_never);
		if (timeout.equals(never)) {
			screenLockTimeout
					.setSummary(R.string.pref_lock_timeout_never_summary);
		} else {
			screenLockTimeout
					.setSummary(R.string.pref_lock_timeout_summary);
		}
	}

	@TargetApi(26)
	private void setupNotificationPreference(SwitchPreference pref,
			String channelId, @StringRes int summary) {
		pref.setWidgetLayoutResource(0);
		pref.setSummary(summary);
		pref.setOnPreferenceClickListener(clickedPref -> {
			Intent intent = new Intent(ACTION_CHANNEL_NOTIFICATION_SETTINGS)
					.putExtra(EXTRA_APP_PACKAGE, getContext().getPackageName())
					.putExtra(EXTRA_CHANNEL_ID, channelId);
			startActivity(intent);
			return true;
		});
	}

	@Override
	public boolean onPreferenceChange(Preference preference, Object newValue) {
		if (preference == language) {
			if (!language.getValue().equals(newValue))
				languageChanged((String) newValue);
			return false;
		} else if (preference == enableBluetooth) {
			boolean btSetting = Boolean.valueOf((String) newValue);
			storeBluetoothSettings(btSetting);
		} else if (preference == torNetwork) {
			int torNetworkSetting = Integer.valueOf((String) newValue);
			storeTorNetworkSetting(torNetworkSetting);
			setTorNetworkSummary(torNetworkSetting);
		} else if (preference == torMobile) {
			boolean torMobileSetting = (Boolean) newValue;
			storeTorMobileSetting(torMobileSetting);
		} else if (preference == screenLock) {
			Settings s = new Settings();
			s.putBoolean(PREF_SCREEN_LOCK, (Boolean) newValue);
			storeSettings(s);
		} else if (preference == screenLockTimeout) {
			Settings s = new Settings();
			String value = (String) newValue;
			s.putInt(PREF_SCREEN_LOCK_TIMEOUT, Integer.valueOf(value));
			storeSettings(s);
			setScreenLockTimeoutSummary(value);
		} else if (preference == notifyLockscreen) {
			Settings s = new Settings();
			s.putBoolean(PREF_NOTIFY_LOCK_SCREEN, (Boolean) newValue);
			storeSettings(s);
		}
		return true;
	}

	private void languageChanged(String newValue) {
		AlertDialog.Builder builder =
				new AlertDialog.Builder(getActivity());
		builder.setTitle(R.string.pref_language_title);
		builder.setMessage(R.string.pref_language_changed);
		builder.setPositiveButton(R.string.sign_out_button,
				(dialogInterface, i) -> {
					language.setValue(newValue);
					Intent intent = new Intent(getContext(),
							NavDrawerActivity.class);
					intent.setFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP);
					intent.putExtra(INTENT_SIGN_OUT, true);
					getActivity().startActivity(intent);
					getActivity().finish();
				});
		builder.setNegativeButton(R.string.cancel, null);
		builder.setCancelable(false);
		builder.show();
	}

	private void storeTorNetworkSetting(int torNetworkSetting) {
		Settings s = new Settings();
		s.putInt(PREF_TOR_NETWORK, torNetworkSetting);
		mergeSettings(s, TOR_NAMESPACE);
	}

	private void storeTorMobileSetting(boolean torMobileSetting) {
		Settings s = new Settings();
		s.putBoolean(PREF_TOR_MOBILE, torMobileSetting);
		mergeSettings(s, TOR_NAMESPACE);
	}

	private void storeBluetoothSettings(boolean btSetting) {
		Settings s = new Settings();
		s.putBoolean(PREF_BT_ENABLE, btSetting);
		mergeSettings(s, BT_NAMESPACE);
	}

	private void storeSettings(Settings s) {
		mergeSettings(s, SETTINGS_NAMESPACE);
	}

	private void mergeSettings(Settings s, String namespace) {
		listener.runOnDbThread(() -> {
			try {
				long start = now();
				settingsManager.mergeSettings(s, namespace);
				logDuration(LOG, "Merging settings", start);
			} catch (DbException e) {
				logException(LOG, WARNING, e);
			}
		});
	}

	@Override
	public void eventOccurred(Event e) {
		if (e instanceof SettingsUpdatedEvent) {
			String namespace = ((SettingsUpdatedEvent) e).getNamespace();
			if (namespace.equals(BT_NAMESPACE)
					|| namespace.equals(TOR_NAMESPACE)
					|| namespace.equals(SETTINGS_NAMESPACE)) {
				LOG.info("Settings updated");
				loadSettings();
			}
		}
	}

}
